from typing import Dict, Optional, Set, Iterable, List

from more_itertools import zip_equal
import torch
from transformers import PreTrainedTokenizer

from adaptor.evaluators.token_classification import TokenClassificationEvaluator
from adaptor.utils import AdaptationDataset


CategoryName = str
GroupName = str
Category = int
Group = Iterable[CategoryName]
CategoryMap = Dict[CategoryName, Category]
GroupMap = Dict[GroupName, Group]

FScore = float


class AggregateMeanFScore(TokenClassificationEvaluator):
    GROUPS: GroupMap = {
        'PER': ('B-PER', 'I-PER', 'B-ORG', 'I-ORG'),
        'LOC': ('B-LOC', 'I-LOC', 'B-ORG', 'I-ORG'),
        'O': ('O', 'B-MISC', 'I-MISC'),
    }

    def __init__(self, category_map: CategoryMap, group_name: Optional[GroupName],
                 *args, **kwargs):
        self.category_map = category_map
        self.group_name = group_name
        super().__init__(*args, **kwargs)

    def __call__(self, model: torch.nn.Module, tokenizer: PreTrainedTokenizer,
                 dataset: AdaptationDataset) -> FScore:
        expected_labels, actual_labels = self._collect_token_predictions(model, dataset)

        if self.group_name is None:
            f_scores = [
                self.get_f_score(self.GROUPS[group_name], expected_labels, actual_labels)
                for group_name
                in self.__class__.get_all_group_names()
            ]
            assert len(f_scores) > 0
            mean_f_score = sum(f_scores) / len(f_scores)
        else:
            group = self.GROUPS[self.group_name]
            mean_f_score = self.get_f_score(group, expected_labels, actual_labels)

        return mean_f_score

    def get_f_score(self, group: Group, expected_labels: List[Category],
                    actual_labels: List[Category]) -> FScore:
        expected_categories: Set[Category] = {self.category_map[category] for category in group}

        true_positives, false_positives, false_negatives = 0, 0, 0
        for expected_label, actual_label in zip_equal(expected_labels, actual_labels):
            if expected_label in expected_categories and actual_label in expected_categories:
                true_positives += 1
            elif expected_label not in expected_categories and actual_label in expected_categories:
                false_positives += 1
            elif expected_label in expected_categories and actual_label not in expected_categories:
                false_negatives += 1

        f_score = true_positives / (true_positives + (0.5 * (false_positives + false_negatives)))
        return f_score

    @classmethod
    def get_all_group_names(cls) -> Iterable[GroupName]:
        for group_name in cls.GROUPS.keys():
            yield group_name

    def __str__(self) -> str:
        if self.group_name is None:
            return 'all'
        else:
            return self.group_name
